﻿using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Globalization;
using Mono.ILASM.yyParser;

namespace Mono.ILASM
{
	public delegate void NewTokenEvent(object sender, NewTokenEventArgs args);

	public class NewTokenEventArgs : EventArgs
	{
		public readonly ILToken Token;

		public NewTokenEventArgs(ILToken token)
		{
			Token = token;
		}
	};
	
	public class ILTokenizer : ITokenStream, yyParser.yyInput
	{
		static readonly string idchars = "_$@?.`";

		public ILTokenizer(StreamReader reader)
		{
			this.reader = new ILReader(reader);
			strBuilder = new StringHelper(this);
			numBuilder = new NumberHelper(this);
			lastToken = ILToken.Invalid;
		}

		ILToken lastToken;
		ILReader reader;
		StringHelper strBuilder;
		NumberHelper numBuilder;
		bool in_byte_array;

		public event NewTokenEvent NewTokenEvent;

		public ILReader Reader => reader;

		public Location Location => reader.Location;

		public bool InByteArray
		{
			get => in_byte_array;
			set => in_byte_array = value;
		}

		public ILToken GetNextToken()
		{
			if (lastToken == ILToken.EOF)
				return lastToken;

			int ch;
			int next;
			ILToken res = ILToken.EOF;

			while ((ch = reader.Read()) != -1)
			{
				// Comments
				if (ch == '/')
				{
					next = reader.Peek();
					if (next == '/')
					{
						// double-slash comment, skip to the end of the line.
						for (reader.Read(); next != -1 && next != '\n'; next = reader.Read())
							;
						continue;
					}
					else if (next == '*')
					{
						reader.Read();
						for (next = reader.Read(); next != -1; next = reader.Read())
						{
							if (next == '*' && reader.Peek() == '/')
							{
								reader.Read();
								goto end;
							}
						}
					end:
						continue;
					}
				}

				// HEXBYTES are flagged by the parser otherwise it is
				// impossible to figure them out
				if (in_byte_array)
				{
					string hx = String.Empty;

					if (Char.IsWhiteSpace((char)ch))
						continue;

					if (ch == ')')
					{
						res = ILToken.CloseParens;
						break;
					}

					if (!is_hex(ch))
						throw new ILTokenizingException(reader.Location, ((char)ch).ToString());
					hx += (char)ch;
					if (is_hex(reader.Peek()))
						hx += (char)reader.Read();
					else if (!Char.IsWhiteSpace((char)reader.Peek()) && reader.Peek() != ')')
						throw new ILTokenizingException(reader.Location, ((char)reader.Peek()).ToString());

					res = new ILToken(Token.HEXBYTE, Byte.Parse(hx, NumberStyles.HexNumber));

					while (Char.IsWhiteSpace((char)reader.Peek()))
						reader.Read();
					break;
				}

				// Ellipsis
				if (ch == '.' && reader.Peek() == '.')
				{
					reader.MarkLocation();
					int ch2 = reader.Read();
					if (reader.Peek() == '.')
					{
						res = ILToken.Ellipsis;
						reader.Read();
						break;
					}
					reader.Unread(ch2);
					reader.RestoreLocation();
				}

				if (ch == '.' || ch == '#')
				{
					next = reader.Peek();
					if (ch == '.' && Char.IsDigit((char)next))
					{
						numBuilder.Start(ch);
						reader.Unread(ch);
						numBuilder.Build();
						if (numBuilder.ResultToken != ILToken.Invalid)
						{
							res = numBuilder.ResultToken;
							break;
						}
					}
					else
					{
						if (strBuilder.Start(next) && strBuilder.TokenId == Token.ID)
						{
							reader.MarkLocation();
							string dirBody = strBuilder.Build();
							string dir = new string((char)ch, 1) + dirBody;
							if (ILTables.Directives.TryGetValue(dir, out ILToken tok))
							{
								res = tok;
							}
							else
							{
								reader.Unread(dirBody.ToCharArray());
								reader.RestoreLocation();
								res = ILToken.Dot;
							}
						}
						else
						{
							res = ILToken.Dot;
						}
						break;
					}
				}

				// Numbers && Hexbytes
				if (numBuilder.Start(ch))
				{
					if ((ch == '-') && !(Char.IsDigit((char)reader.Peek())))
					{
						res = ILToken.Dash;
						break;
					}
					else
					{
						reader.Unread(ch);
						numBuilder.Build();
						if (numBuilder.ResultToken != ILToken.Invalid)
						{
							res = numBuilder.ResultToken;
							break;
						}
					}
				}

				// Punctuation
				ILToken punct = ILToken.GetPunctuation(ch);
				if (punct != null)
				{
					if (punct == ILToken.Colon && reader.Peek() == ':')
					{
						reader.Read();
						res = ILToken.DoubleColon;
					}
					else
					{
						res = punct;
					}
					break;
				}

				// ID | QSTRING | SQSTRING | INSTR_* | KEYWORD
				if (strBuilder.Start(ch))
				{
					reader.Unread(ch);
					string val = strBuilder.Build();
					if (strBuilder.TokenId == Token.ID)
					{
						ILToken opcode;
						next = reader.Peek();
						if (next == '.')
						{
							reader.MarkLocation();
							reader.Read();
							next = reader.Peek();
							if (IsIdChar((char)next))
							{
								string opTail = BuildId();
								string full_str = String.Format("{0}.{1}", val, opTail);

								if (ILTables.Instr.TryGetValue(full_str, out opcode))
									res = opcode;
								else
									res = new ILToken(Token.COMP_NAME, full_str);
								break;
							}
							else if (Char.IsWhiteSpace((char)next))
							{
								// Handle 'tail.' and 'unaligned.'
								if (ILTables.Instr.TryGetValue(val + ".", out opcode))
								{
									res = opcode;
									break;
								}
								// Let the parser handle the dot
								reader.Unread('.');
							}
						}
						if (ILTables.Instr.TryGetValue(val, out opcode))
						{
							res = opcode;
							break;
						}
						if (ILTables.Keywords.TryGetValue(val, out opcode))
						{
							res = opcode;
							break;
						}
					}

					res = new ILToken(strBuilder.TokenId, val);
					break;
				}
			}

			OnNewToken(res);
			lastToken = res;
			return res;
		}

		public ILToken NextToken => GetNextToken();

		public ILToken LastToken => lastToken;

		bool is_hex(int e) => (e >= '0' && e <= '9') || (e >= 'A' && e <= 'F') || (e >= 'a' && e <= 'f');

		static bool IsIdStartChar(char ch) => Char.IsLetter(ch) || idchars.IndexOf(ch) != -1;

		static bool IsIdChar(char ch) => Char.IsLetterOrDigit(ch) || idchars.IndexOf(ch) != -1;

		string BuildId()
		{
			StringBuilder idsb = new StringBuilder();
			int ch, last;

			last = -1;
			while ((ch = reader.Read()) != -1)
			{
				if (IsIdChar((char)ch) || ch == '.')
				{
					idsb.Append((char)ch);
				}
				else
				{
					reader.Unread(ch);
					// Never end an id on a DOT
					if (last == '.')
					{
						reader.Unread(last);
						idsb.Length -= 1;
					}
					break;
				}
				last = ch;
			}

			return idsb.ToString();
		}

		void OnNewToken(ILToken token)
		{
			NewTokenEvent?.Invoke(this, new NewTokenEventArgs(token));
		}

		bool yyInput.advance()
		{
			return GetNextToken() != ILToken.EOF;
		}

		int yyInput.token()
		{
			return lastToken.TokenId;
		}

		object yyInput.value()
		{
			return lastToken.Value;
		}
	};
}
